/************************************************************************
 *
 * Copyright (C) 2023-2024 IRCAD France
 *
 * This file is part of Sight.
 *
 * Sight is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Sight is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with Sight. If not, see <https://www.gnu.org/licenses/>.
 *
 ***********************************************************************/

#pragma once

#include <cppunit/extensions/HelperMacros.h>

// cspell:ignore nvjpeg

namespace sight::io::bitmap::ut
{

class reader_writer_test : public CPPUNIT_NS::TestFixture
{
public:

    CPPUNIT_TEST_SUITE(reader_writer_test);
    CPPUNIT_TEST(grayscale_uint_8_test);
    CPPUNIT_TEST(grayscale_uint_16_test);
    CPPUNIT_TEST(grayscale_int_8_test);
    CPPUNIT_TEST(grayscale_int_16_test);
    CPPUNIT_TEST(rgb_uint_8_test);
    CPPUNIT_TEST(rgb_uint_16_test);
    CPPUNIT_TEST(rgb_int_8_test);
    CPPUNIT_TEST(rgb_int_16_test);
    CPPUNIT_TEST(rgba_uint_8_test);
    CPPUNIT_TEST(rgba_uint_16_test);
    CPPUNIT_TEST(rgba_int_8_test);
    CPPUNIT_TEST(rgba_int_16_test);
    CPPUNIT_TEST_SUITE_END();

public:

    void setUp() override;
    void tearDown() override;

    static void grayscale_uint_8_test();
    static void grayscale_uint_16_test();
    static void grayscale_int_8_test();
    static void grayscale_int_16_test();
    static void rgb_uint_8_test();
    static void rgb_uint_16_test();
    static void rgb_int_8_test();
    static void rgb_int_16_test();
    static void rgba_uint_8_test();
    static void rgba_uint_16_test();
    static void rgba_int_8_test();
    static void rgba_int_16_test();
};

} // namespace sight::io::bitmap::ut
