/************************************************************************
 *
 * Copyright (C) 2009-2023 IRCAD France
 * Copyright (C) 2012-2015 IHU Strasbourg
 *
 * This file is part of Sight.
 *
 * Sight is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Sight is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with Sight. If not, see <https://www.gnu.org/licenses/>.
 *
 ***********************************************************************/

// cspell:ignore NOLINT

#include "io/vtk/helper/vtk_lambda_command.hpp"

namespace sight::io::vtk::helper
{

//------------------------------------------------------------------------------

void vtk_lambda_command::Execute(vtkObject* _caller, unsigned long _event, void* _call_data) // NOLINT(google-runtime-int)
{
    if(m_callback)
    {
        m_callback(_caller, _event, _call_data);
    }
}

} // namespace sight::io::vtk::helper
