/************************************************************************
 *
 * Copyright (C) 2016-2024 IRCAD France
 * Copyright (C) 2016-2020 IHU Strasbourg
 *
 * This file is part of Sight.
 *
 * Sight is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Sight is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with Sight. If not, see <https://www.gnu.org/licenses/>.
 *
 ***********************************************************************/

#pragma once

#include <sight/ui/qt/config.hpp>

#include "ui/qt/container/widget.hpp"

#include <core/base.hpp>

#include <ui/__/builder/slideview.hpp>
#include <ui/__/container/widget.hpp>

namespace sight::ui::qt::builder
{

/// This builder creates a slide widget. The slide widget can be displayed over all others widgets.
class SIGHT_UI_QT_CLASS_API slideview : public ui::builder::slideview
{
public:

    SIGHT_DECLARE_CLASS(
        slideview,
        ui::builder::slideview,
        ui::factory::make<slideview>
    )

    /// Destroyes the slide builder.
    SIGHT_UI_QT_API ~slideview() override = default;

    /// Create the Qt container @ref ui::qt::widget::slide_bar.
    SIGHT_UI_QT_API void create_container(ui::container::widget::sptr _parent) override;

    /// Destroyes the container.
    SIGHT_UI_QT_API void destroy_container() override;

protected:

    /// Contains the generic parent container.
    ui::qt::container::widget::sptr m_parent;
};

} // namespace sight::ui::qt::builder
