/************************************************************************
 *
 * Copyright (C) 2009-2023 IRCAD France
 * Copyright (C) 2012-2019 IHU Strasbourg
 *
 * This file is part of Sight.
 *
 * Sight is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Sight is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with Sight. If not, see <https://www.gnu.org/licenses/>.
 *
 ***********************************************************************/

#include "io/igtl/detail/converter/scalar_converter.hpp"

#include "io/igtl/detail/data_converter.hpp"

#include <igtlStringMessage.h>

namespace sight::io::igtl::detail::converter
{

CONVERTER_REGISTER_MACRO(io::igtl::detail::converter::int_converter);
CONVERTER_REGISTER_MACRO(io::igtl::detail::converter::float_converter);

//-----------------------------------------------------------------------------

} // namespace sight::io::igtl::detail::converter
