/************************************************************************
 *
 * Copyright (C) 2018-2024 IRCAD France
 * Copyright (C) 2018-2020 IHU Strasbourg
 *
 * This file is part of Sight.
 *
 * Sight is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Sight is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with Sight. If not, see <https://www.gnu.org/licenses/>.
 *
 ***********************************************************************/

#pragma once

#include "sight/tuto13_mesh_generator_cpp/config.hpp"

#include <core/runtime/plugin.hpp>

namespace tuto13_mesh_generator_cpp
{

/// This class is started when the module is loaded.
class SIGHT_TUTO13_MESH_GENERATOR_CPP_CLASS_API plugin final : public sight::core::runtime::plugin
{
public:

    /// Destroys the plugin.
    SIGHT_TUTO13_MESH_GENERATOR_CPP_API ~plugin() noexcept override;

    /// Starts the plugin, does nothing here.
    SIGHT_TUTO13_MESH_GENERATOR_CPP_API void start() override;

    /// Stops the plugin, does nothing here.
    SIGHT_TUTO13_MESH_GENERATOR_CPP_API void stop() noexcept override;
};

} // namespace tuto13_mesh_generator_cpp.
