/************************************************************************
 *
 * Copyright (C) 2021-2024 IRCAD France
 *
 * This file is part of Sight.
 *
 * Sight is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Sight is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with Sight. If not, see <https://www.gnu.org/licenses/>.
 *
 ***********************************************************************/

#pragma once

#include <sight/ui/test/config.hpp>

#include "tester.hpp"

#include <core/runtime/profile/profile.hpp>

#include <cppunit/extensions/HelperMacros.h>

#include <QEvent>
#include <QObject>

namespace sight::ui::test
{

class SIGHT_UI_TEST_CLASS_API base : public CPPUNIT_NS::TestFixture
{
public:

    SIGHT_UI_TEST_API void setUp() override;

protected:

    SIGHT_UI_TEST_API void start(
        const std::string& _test_name,
        std::function<void(tester&)> _test,
        bool _verbose_mode = false
    );

    SIGHT_UI_TEST_API virtual std::filesystem::path get_profile_path() = 0;

    SIGHT_UI_TEST_API static void compare_images(const std::filesystem::path& _a, const std::filesystem::path& _b);

private:

    sight::core::runtime::profile::sptr m_profile;
};

} // namespace sight::ui::testCore
